# from tikz2graphml import grammar
# from tikz2graphml import pyyed

import os
import antlr4
import logging
from tikz2graphml.grammar.TikzLexer import TikzLexer
from tikz2graphml.grammar.TikzParser import TikzParser
from tikz2graphml.TikzErrorListener import TikzErrorListener
from tikz2graphml.CustomTikzListener import CustomTikzListener
from tikz2graphml.extradeCodeInsideTikzAndUnrollForeach import getCodeInsideTIKZAfterUnrolling

logger = logging.getLogger()

"""
Higher level class to handle conversion of Tikz graph to GraphML format
"""
class ParseTikz:
    def printContents(self, value):
        print("===================================\n")
        for i, line in enumerate(value.split('\n')):
            print(i+1, ": ", line)
        print("===================================\n")

    """
    Main conversion function
    Arguments:-
    scalingFactor:  scale value provided by the user to make graph look reasonable
    logLevel: Toggle verbosity
    inputFilename: path of the input Tikz File
    directory: path of the output directory to save the generated GraphML file
    """
    def run(self, scalingFactor: float, inputFilename: str, prefix: str, directory: str):

        if not prefix:
            prefix = os.path.basename(os.path.splitext(prefix)[0])

        if not os.path.exists(directory):
            os.makedirs(directory)

        for value in getCodeInsideTIKZAfterUnrolling(inputFilename):
            self.printContents(value)
            input_stream = antlr4.InputStream(value)
            lexer = TikzLexer(input_stream)
            stream = antlr4.CommonTokenStream(lexer)
            parser = TikzParser(stream)
            parser.addErrorListener(TikzErrorListener())
            tree = parser.begin()

            # we save file as filename_t_{n}_graph.graphml
            # Getting next available output file path
            j = 0
            while(os.path.exists(directory + "/" + prefix + "_" + str(j) + "_graph.graphml")):
                j += 1
            outputFilename = directory + "/" + prefix + "_" + str(j) + "_graph.graphml"
            htmlChat = CustomTikzListener(inputFilename, outputFilename, scalingFactor)
            walker = antlr4.ParseTreeWalker()
            walker.walk(htmlChat, tree)



import re
import logging
import tikz2graphml.filterGraphml as filterGraphml
from tikz2graphml.grammar.TikzParser import TikzParser

logger = logging.getLogger()

supportedNodeTags = ["nodeID", "X", "Y", "label", "height", "width", "inner_sep", "fill", "edge_color",
                     "scale", "shape", "regular_polygon_sides", "rotate", "auto"]

def filterOutNotSupportedNodeTags(propertyDict):
    for k in list(propertyDict):
        if k not in supportedNodeTags:
            del propertyDict[k]

# Handle Parsing of label strings obtained from Antlr
def parseLabelValue(labelstr):

    if labelstr[0] != '{':
        raise Exception("Expected '{' at the beginning of label. Label starts with incorrect bracket. Got " + labelstr)

    new_label_str = labelstr[1:]
    slice_index = -1
    for i in range(len(new_label_str)-1, -1,-1):
        if new_label_str[i] == '}':
            slice_index = i
            break
        else:
            continue

    if slice_index == -1:
        raise Exception("Expected '};' at the end of label. Got " + labelstr)

    return new_label_str[:slice_index]

# Number handling. Also handles floats
def handleNumbers(input):
    m = re.search('^\s*([0-9/*-+.]+)\s*(?:(pt|cm))?\s*$', input)
    try:
        if m and len(m.group(1)) > 0 and m.group(1) != ".":
            if m.group(2) and m.group(2) == "cm":
                return float(eval(m.group(1)))
            return float(eval(m.group(1)))
    except:
        raise Exception("Cannot Evaluate Math Expression {}".format(input))

# Parses a list of property strings and retures a dictionary of those properties
def handleProperties(ctx: TikzParser.PropertiesContext):
    # properties: individualProperty
    if ctx.getChildCount() == 1:
        k, v = handleIndividualProperty(ctx.getTypedRuleContexts(TikzParser.IndividualPropertyContext)[0])
        if k is None:
            return {}
        else:
            return {k: v}

    # properties: individualProperty ',' properties
    else:
        k, v = handleIndividualProperty(ctx.getTypedRuleContext(TikzParser.IndividualPropertyContext, 0))
        listProperties = handleProperties(ctx.getTypedRuleContext(TikzParser.PropertiesContext, 0))

        # If same property is present, the later one is given precedence
        if k not in listProperties:
            listProperties[k] = v

        return listProperties

def handleIndividualProperty(ctx: TikzParser.IndividualPropertyContext):
    key = ""
    value = ""
    currentValue = ""
    for child in ctx.children:
        if child.getText() != "=":
            currentValue += child.getText() + " "
        else:
            key = currentValue
            currentValue = ""
    value = currentValue
    # property of Key Value of format "x = y"
    if len(key) > 0:
        k, v = filterGraphml.identifyKeyValueProperty(key, value)
    # individual property of format "x"
    else:
        k, v = filterGraphml.identifyIndividualProperty(value)

    return k, v



from antlr4.error.ErrorListener import ErrorListener


class TikzErrorListener(ErrorListener):

    def __init__(self):
        super(TikzErrorListener, self).__init__()

    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
        raise Exception("""\n\nSyntax Error in/around the following line. :
            \n\tLine Number : {}
            \n\tColumn : {}
            \n\tOffending Symbol : {}
            \n\tMessage : {}
            \n\tP.S. Look at text just before the offending symbol.\n\n""".format(line, column, offendingSymbol.text, msg))


import sys
import xml.etree.cElementTree as ET

node_shapes = ["rectangle", "rectangle3d", "roundrectangle", "diamond", "ellipse",
               "fatarrow", "fatarrow2", "hexagon", "octagon", "parallelogram",
               "parallelogram2", "star5", "star6", "star6", "star8", "trapezoid",
               "trapezoid2", "triangle", "trapezoid2", "triangle"]

line_types = ["line", "dashed", "dotted", "dashed_dotted"]
font_styles = ["plain", "bold", "italic", "bolditalic"]

arrow_types = ["none", "standard", "white_delta", "diamond", "white_diamond", "short",
               "plain", "concave", "convex", "circle", "transparent_circle", "dash",
               "skewed_dash", "t_shape", "crows_foot_one_mandatory",
               "crows_foot_many_mandatory", "crows_foot_many_optional", "crows_foot_one",
               "crows_foot_many", "crows_foot_optional"]


class Group:
    def __init__(self, group_id, parent_graph, label=None, shape="rectangle",
                 closed="false", font_family="Dialog", underlined_text="false",
                 font_style="plain", font_size="12", fill="#FFCC00", transparent="false",
                 edge_color="#000000", edge_type="line", edge_width="1.0", height=False,
                 width=False, x=False, y=False):
        self.label = label
        if label is None:
            self.label = group_id

        self.group_id = group_id
        self.nodes = {}
        self.parent_graph = parent_graph

        # node shape
        if shape not in node_shapes:
            raise RuntimeWarning("Node shape %s not recognised" % shape)
        self.shape = shape

        self.closed = closed

        # label formatting options
        self.font_family = font_family
        self.underlined_text = underlined_text

        if font_style not in font_styles:
            raise RuntimeWarning("Font style %s not recognised" % font_style)

        self.font_style = font_style
        self.font_size = font_size

        self.fill = fill
        self.transparent = transparent

        self.geom = {}
        if height:
            self.geom["height"] = height
        if width:
            self.geom["width"] = width
        if x:
            self.geom["x"] = x
        if y:
            self.geom["y"] = y

        self.edge_color = edge_color
        self.edge_width = edge_width

        if edge_type not in line_types:
            raise RuntimeWarning("Edge type %s not recognised" % edge_type)

        self.edge_type = edge_type

    def add_node(self, node_name, **kwargs):
        if node_name in self.nodes.keys():
            raise RuntimeWarning("Node %s already exists" % node_name)

        self.nodes[node_name] = Node(node_name, **kwargs)
        self.parent_graph.nodes_in_groups.append(node_name)

    def convert(self):
        node = ET.Element("node", id=self.group_id)
        node.set("yfiles.foldertype", "group")
        data = ET.SubElement(node, "data", key="data_node")

        # node for group
        pabn = ET.SubElement(data, "y:ProxyAutoBoundsNode")
        r = ET.SubElement(pabn, "y:Realizers", active="0")
        group_node = ET.SubElement(r, "y:GroupNode")

        if self.geom:
            ET.SubElement(group_node, "y:Geometry", **self.geom)

        ET.SubElement(group_node, "y:Fill", color=self.fill, transparent=self.transparent)

        ET.SubElement(group_node, "y:BorderStyle", color=self.edge_color,
                      type=self.edge_type, width=self.edge_width)

        label = ET.SubElement(group_node, "y:NodeLabel", modelName="internal",
                              modelPosition="t",
                              fontFamily=self.font_family, fontSize=self.font_size,
                              underlinedText=self.underlined_text,
                              fontStyle=self.font_style)
        label.text = self.label

        ET.SubElement(group_node, "y:Shape", type=self.shape)

        ET.SubElement(group_node, "y:State", closed=self.closed)

        graph = ET.SubElement(node, "graph", edgedefault="directed", id=self.group_id)

        for node_id in self.nodes:
            n = self.nodes[node_id].convert()
            graph.append(n)

        return node
        # ProxyAutoBoundsNode crap just draws bar at top of group


class Node:
    def __init__(self, node_name, label=None, shape="rectangle", font_family="Dialog",
                 underlined_text="false", font_style="plain", font_size="12",
                 shape_fill="#FF0000", transparent="false", edge_color="#000000",
                 edge_type="line", edge_width="1.0", height=False, width=False, x=False,
                 y=False, node_type="ShapeNode", UML=False):

        self.label = label
        # if label is None:
        #     self.label = node_name

        self.node_name = node_name

        self.node_type = node_type
        self.UML = UML

        # node shape
        if shape not in node_shapes:
            raise RuntimeWarning("Node shape %s not recognised" % shape)

        self.shape = shape

        # label formatting options
        self.font_family = font_family
        self.underlined_text = underlined_text

        if font_style not in font_styles:
            raise RuntimeWarning("Font style %s not recognised" % font_style)

        self.font_style = font_style
        self.font_size = font_size

        # shape fill
        self.shape_fill = shape_fill
        self.transparent = transparent

        # edge options
        self.edge_color = edge_color
        self.edge_width = edge_width

        if edge_type not in line_types:
            raise RuntimeWarning("Edge type %s not recognised" % edge_type)

        self.edge_type = edge_type

        # geometry
        self.geom = {}
        if height:
            self.geom["height"] = height
        if width:
            self.geom["width"] = width
        if x:
            self.geom["x"] = x
        if y:
            self.geom["y"] = y

    def convert(self):

        node = ET.Element("node", id=str(self.node_name))
        data = ET.SubElement(node, "data", key="data_node")
        shape = ET.SubElement(data, "y:" + self.node_type)

        if self.geom:
            ET.SubElement(shape, "y:Geometry", **self.geom)
        # <y:Geometry height="30.0" width="30.0" x="475.0" y="727.0"/>

        if self.shape_fill is None:
            ET.SubElement(shape, "y:Fill", hasColor="false",
                        transparent=self.transparent)
        else:
            ET.SubElement(shape, "y:Fill", color=self.shape_fill,
                      transparent=self.transparent)

        if self.edge_color is None:
            ET.SubElement(shape, "y:BorderStyle", hasColor="false", type=self.edge_type,
                        width=self.edge_width)
        else:
            ET.SubElement(shape, "y:BorderStyle", color=self.edge_color, type=self.edge_type,
                        width=self.edge_width)

        label = ET.SubElement(shape, "y:NodeLabel", fontFamily=self.font_family,
                              fontSize=self.font_size,
                              underlinedText=self.underlined_text,
                              fontStyle=self.font_style)
        label.text = self.label

        ET.SubElement(shape, "y:Shape", type=self.shape)

        if self.UML:
            UML = ET.SubElement(shape, "y:UML")

            attributes = ET.SubElement(UML, "y:AttributeLabel", type=self.shape)
            attributes.text = self.UML["attributes"]

            methods = ET.SubElement(UML, "y:MethodLabel", type=self.shape)
            methods.text = self.UML["methods"]

            stereotype = self.UML["stereotype"] if "stereotype" in self.UML else ""
            UML.set("stereotype", stereotype)

        return node


class Edge:
    def __init__(self, node1, node2, label="", arrowhead="none", arrowfoot="none",
                 color="#000000", line_type="line", width="1.0"):
        self.node1 = node1
        self.node2 = node2
        self.edge_id = "%s_%s" % (node1, node2)

        self.label = label

        if arrowhead not in arrow_types:
            raise RuntimeWarning("Arrowhead type %s not recognised" % arrowhead)

        self.arrowhead = arrowhead

        if arrowfoot not in arrow_types:
            raise RuntimeWarning("Arrowhead type %s not recognised" % arrowfoot)

        self.arrowfoot = arrowfoot

        if line_type not in line_types:
            raise RuntimeWarning("Edge type %s not recognised" % line_type)

        self.line_type = line_type

        self.color = color
        self.width = width

    def convert(self):
        edge = ET.Element("edge", id=str(self.edge_id), source=str(self.node1), target=str(self.node2))
        data = ET.SubElement(edge, "data", key="data_edge")
        pl = ET.SubElement(data, "y:PolyLineEdge")

        ET.SubElement(pl, "y:Arrows", source=self.arrowfoot, target=self.arrowhead)
        ET.SubElement(pl, "y:LineStyle", color=self.color, type=self.line_type,
                      width=self.width)

        if self.label:
            ET.SubElement(pl, "y:EdgeLabel").text = self.label

        return edge


class Graph:
    def __init__(self, directed="directed", graph_id="G"):

        self.nodes_in_groups = []
        self.nodes = {}
        self.edges = {}
        self.orderOfArrivalNodes = []
        self.orderOfArrivalEdges = []
        self.orderOfArrivalGroups = []

        self.directed = directed
        self.graph_id = graph_id

        self.groups = {}

        self.graphml = ""

    def construct_graphml(self):
        # xml = ET.Element("?xml", version="1.0", encoding="UTF-8", standalone="no")

        graphml = ET.Element("graphml", xmlns="http://graphml.graphdrawing.org/xmlns")
        graphml.set("xmlns:java", "http://www.yworks.com/xml/yfiles-common/1.0/java")
        graphml.set("xmlns:sys",
                    "http://www.yworks.com/xml/yfiles-common/markup/primitives/2.0")
        graphml.set("xmlns:x", "http://www.yworks.com/xml/yfiles-common/markup/2.0")
        graphml.set("xmlns:xsi", "http://www.w3.org/2001/XMLSchema-instance")
        graphml.set("xmlns:y", "http://www.yworks.com/xml/graphml")
        graphml.set("xmlns:yed", "http://www.yworks.com/xml/yed/3")
        graphml.set("xsi:schemaLocation",
                    "http://graphml.graphdrawing.org/xmlns http://www.yworks.com/xml/schema/graphml/1.1/ygraphml.xsd")

        node_key = ET.SubElement(graphml, "key", id="data_node")
        node_key.set("for", "node")
        node_key.set("yfiles.type", "nodegraphics")

        edge_key = ET.SubElement(graphml, "key", id="data_edge")
        edge_key.set("for", "edge")
        edge_key.set("yfiles.type", "edgegraphics")

        graph = ET.SubElement(graphml, "graph", edgedefault=self.directed,
                              id=self.graph_id)

        for node_id in self.orderOfArrivalNodes:
            node = self.nodes[node_id].convert()
            graph.append(node)

        for group_id in self.orderOfArrivalGroups:
            node = self.groups[group_id].convert()
            graph.append(node)

        for edge_id in self.orderOfArrivalEdges:
            edge = self.edges[edge_id].convert()
            graph.append(edge)

        self.graphml = graphml

    def write_graph(self, filename):
        self.construct_graphml()
        tree = ET.ElementTree(self.graphml)
        tree.write(filename)

    def get_graph(self):
        self.construct_graphml()
        # Py2/3 sigh.
        if sys.version_info.major < 3:
            return ET.tostring(self.graphml, encoding='UTF-8')
        else:
            return ET.tostring(self.graphml, encoding='UTF-8').decode()

    def add_node(self, node_name, **kwargs):
        if node_name in self.nodes.keys():
            raise RuntimeWarning("Node %s already exists" % node_name)
        
        self.nodes[node_name] = Node(node_name, **kwargs)
        self.orderOfArrivalNodes.append(node_name)

    def add_edge(self, node1, node2, label="", arrowhead="convex", arrowfoot="none",
                 color="#000000", line_type="line",
                 width="1.0"):
        # pass node names, not actual node objects

        existing_entities = self.nodes_in_groups
        existing_entities.extend(self.nodes.keys())
        existing_entities.extend(self.groups.keys())

        if node1 not in existing_entities:
            self.nodes[node1] = Node(node1)

        if node2 not in existing_entities:
            self.nodes[node2] = Node(node2)

        edge = Edge(node1, node2, label, arrowhead, arrowfoot, color, line_type, width)
        self.edges[edge.edge_id] = edge
        self.orderOfArrivalEdges.append(edge.edge_id)

    def add_group(self, group_id, **kwargs):
        self.groups[group_id] = Group(group_id, self, **kwargs)
        self.orderOfArrivalGroups.append(group_id)
        return self.groups[group_id]



import logging
from colour import Color

logger = logging.getLogger()

ValidNodeShapes = ["rectangle", "circle", "diamond", "ellipse"]
ValidEdgeDirections = ["->", "<-", "-!-", "--", "<->", "-"]  #- means no edge. So no need to check for its validity as that is default

"""
A list of functions which try to check the validity of each attribute
and optionally defining a default value wherever required
"""

def isValidColor(value):
    # If value contains ! then it is definatly color (not seem ! in any other variable/property)
    if value.__contains__("!"):
        return True
    # Check if we can get color object of given value or not
    try:
        Color(value)
        return True
    except:
        return False

def isValidShape(value):
    return value in ValidNodeShapes

def isValidDirection(value):
    return value in ValidEdgeDirections

# Refer http://www.texample.net/tikz/examples/node-shapes/
# for Tikz shapes
def identifyShape(value):
    return value if value in ValidNodeShapes else "rectangle"

def isValidThickness(value):
    return value.lower() in ["thick", "thin", "ultra thich"]

def isValidEdgeStyle(value):
    return value.lower() in ["dash", "solid", "dotted" ]

def identifyIndividualProperty(value):
    value = value.strip().lower()
    if isValidColor(value):
        return ("fill", value)
    elif isValidShape(value):
        return ("shape", value)
    elif isValidDirection(value):
        return ("direction", value)
    elif isValidThickness(value):
        return ("width", value)
    elif isValidEdgeStyle(value):
        return ("line_type", value)

    # Add other checks here

    # We could not identify what property it was so return none
    logger.warning("Got Unhandled Property, Ignoring it for now ({})".format(value))
    return (None, None)

def identifyKeyValueProperty(key: str, value: str):
    key, value = key.strip().lower(), value.strip().lower()

    if key == "shape":
        return (key, identifyShape(value))

    elif key in ["fill", "rotate", "label"]:
        return (key, value)

    elif key == "inner sep":
        return ("inner_sep", value)

    elif key == "direction":
        return ("direction", value)

    elif key == "thickness" or key == "width":
        return ("width", value)

    elif key == "edgeStyle":
        return ("edgeStyle", value)

    elif key == "edge label":
        return ("label", value)

    elif key == "scale":
        return ("scale", value)

    elif key == "auto":
        return ("auto", value)

    # Add other checks here

    logger.warning("Got Unhandled Property, Ignoring it for now ({},{})".format(key, value))
    return None, None



# Generated from tikz2graphml/grammar/Tikz.g4 by ANTLR 4.7.2
# encoding: utf-8
from antlr4 import *
from io import StringIO
from typing.io import TextIO
import sys


def serializedATN():
    with StringIO() as buf:
        buf.write("\3\u608b\ua72a\u8133\ub9ed\u417c\u3be7\u7786\u5964\3\34")
        buf.write("\u00ca\4\2\t\2\4\3\t\3\4\4\t\4\4\5\t\5\4\6\t\6\4\7\t\7")
        buf.write("\4\b\t\b\4\t\t\t\4\n\t\n\4\13\t\13\4\f\t\f\4\r\t\r\4\16")
        buf.write("\t\16\4\17\t\17\4\20\t\20\4\21\t\21\3\2\3\2\3\2\7\2&\n")
        buf.write("\2\f\2\16\2)\13\2\3\2\3\2\3\2\3\3\3\3\3\3\3\3\3\3\3\3")
        buf.write("\3\3\3\3\5\3\66\n\3\3\4\3\4\3\4\3\4\3\4\3\4\3\4\3\4\3")
        buf.write("\4\3\4\3\4\3\4\3\4\3\4\3\4\3\4\3\4\3\4\3\4\3\4\3\4\3\4")
        buf.write("\3\4\3\4\3\4\3\4\5\4R\n\4\3\5\3\5\3\5\3\5\3\6\3\6\3\6")
        buf.write("\3\6\3\6\5\6]\n\6\3\7\3\7\3\7\3\7\5\7c\n\7\3\b\3\b\5\b")
        buf.write("g\n\b\3\b\3\b\5\bk\n\b\3\t\3\t\3\t\3\t\3\t\3\t\3\t\3\t")
        buf.write("\3\t\3\n\3\n\5\nx\n\n\3\n\3\n\5\n|\n\n\3\13\3\13\5\13")
        buf.write("\u0080\n\13\3\13\3\13\5\13\u0084\n\13\3\f\3\f\3\f\3\f")
        buf.write("\3\f\3\f\3\f\3\f\3\f\3\f\3\f\5\f\u0091\n\f\3\f\3\f\3\f")
        buf.write("\7\f\u0096\n\f\f\f\16\f\u0099\13\f\3\r\3\r\5\r\u009d\n")
        buf.write("\r\3\r\3\r\5\r\u00a1\n\r\3\16\3\16\3\16\3\16\3\16\5\16")
        buf.write("\u00a8\n\16\3\17\6\17\u00ab\n\17\r\17\16\17\u00ac\3\17")
        buf.write("\3\17\6\17\u00b1\n\17\r\17\16\17\u00b2\3\17\6\17\u00b6")
        buf.write("\n\17\r\17\16\17\u00b7\5\17\u00ba\n\17\3\20\3\20\3\20")
        buf.write("\3\20\3\20\3\20\3\20\3\20\3\20\3\20\5\20\u00c6\n\20\3")
        buf.write("\21\3\21\3\21\2\3\26\22\2\4\6\b\n\f\16\20\22\24\26\30")
        buf.write("\32\34\36 \2\4\3\2\31\32\4\2\17\17\24\24\2\u00d2\2\"\3")
        buf.write("\2\2\2\4\65\3\2\2\2\6Q\3\2\2\2\bS\3\2\2\2\n\\\3\2\2\2")
        buf.write("\fb\3\2\2\2\16j\3\2\2\2\20l\3\2\2\2\22{\3\2\2\2\24\u0083")
        buf.write("\3\2\2\2\26\u0090\3\2\2\2\30\u00a0\3\2\2\2\32\u00a7\3")
        buf.write("\2\2\2\34\u00b9\3\2\2\2\36\u00c5\3\2\2\2 \u00c7\3\2\2")
        buf.write("\2\"#\7\n\2\2#\'\5\24\13\2$&\5\4\3\2%$\3\2\2\2&)\3\2\2")
        buf.write("\2\'%\3\2\2\2\'(\3\2\2\2(*\3\2\2\2)\'\3\2\2\2*+\7\13\2")
        buf.write("\2+,\7\2\2\3,\3\3\2\2\2-.\5\20\t\2./\5\4\3\2/\66\3\2\2")
        buf.write("\2\60\61\5\6\4\2\61\62\5\4\3\2\62\66\3\2\2\2\63\66\5\6")
        buf.write("\4\2\64\66\5\20\t\2\65-\3\2\2\2\65\60\3\2\2\2\65\63\3")
        buf.write("\2\2\2\65\64\3\2\2\2\66\5\3\2\2\2\678\7\r\2\289\5\16\b")
        buf.write("\29:\5\n\6\2:;\7\26\2\2;R\3\2\2\2<=\7\r\2\2=>\5\16\b\2")
        buf.write(">?\5\36\20\2?@\t\2\2\2@A\5\36\20\2AB\7\26\2\2BR\3\2\2")
        buf.write("\2CD\7\r\2\2DE\5\16\b\2EF\5\36\20\2FG\t\2\2\2GH\5\b\5")
        buf.write("\2HI\7\26\2\2IR\3\2\2\2JK\7\r\2\2KL\5\16\b\2LM\5\36\20")
        buf.write("\2MN\t\2\2\2NO\5\30\r\2OP\5 \21\2PR\3\2\2\2Q\67\3\2\2")
        buf.write("\2Q<\3\2\2\2QC\3\2\2\2QJ\3\2\2\2R\7\3\2\2\2ST\7\21\2\2")
        buf.write("TU\t\2\2\2UV\7\22\2\2V\t\3\2\2\2WX\5\f\7\2XY\7\3\2\2Y")
        buf.write("Z\5\n\6\2Z]\3\2\2\2[]\5\f\7\2\\W\3\2\2\2\\[\3\2\2\2]\13")
        buf.write("\3\2\2\2^c\5\36\20\2_`\7\21\2\2`a\t\2\2\2ac\7\22\2\2b")
        buf.write("^\3\2\2\2b_\3\2\2\2c\r\3\2\2\2df\7\4\2\2eg\5\32\16\2f")
        buf.write("e\3\2\2\2fg\3\2\2\2gh\3\2\2\2hk\7\5\2\2ik\3\2\2\2jd\3")
        buf.write("\2\2\2ji\3\2\2\2k\17\3\2\2\2lm\7\f\2\2mn\5\30\r\2no\5")
        buf.write("\22\n\2op\5\30\r\2pq\7\16\2\2qr\5\36\20\2rs\5\30\r\2s")
        buf.write("t\5 \21\2t\21\3\2\2\2uw\7\21\2\2vx\t\2\2\2wv\3\2\2\2w")
        buf.write("x\3\2\2\2xy\3\2\2\2y|\7\22\2\2z|\3\2\2\2{u\3\2\2\2{z\3")
        buf.write("\2\2\2|\23\3\2\2\2}\177\7\4\2\2~\u0080\5\26\f\2\177~\3")
        buf.write("\2\2\2\177\u0080\3\2\2\2\u0080\u0081\3\2\2\2\u0081\u0084")
        buf.write("\7\5\2\2\u0082\u0084\3\2\2\2\u0083}\3\2\2\2\u0083\u0082")
        buf.write("\3\2\2\2\u0084\25\3\2\2\2\u0085\u0086\b\f\1\2\u0086\u0087")
        buf.write("\7\20\2\2\u0087\u0088\t\2\2\2\u0088\u0089\7\6\2\2\u0089")
        buf.write("\u008a\7\7\2\2\u008a\u008b\7\23\2\2\u008b\u008c\7\b\2")
        buf.write("\2\u008c\u008d\5\32\16\2\u008d\u008e\7\t\2\2\u008e\u0091")
        buf.write("\3\2\2\2\u008f\u0091\5\32\16\2\u0090\u0085\3\2\2\2\u0090")
        buf.write("\u008f\3\2\2\2\u0091\u0097\3\2\2\2\u0092\u0093\f\5\2\2")
        buf.write("\u0093\u0094\7\24\2\2\u0094\u0096\5\26\f\6\u0095\u0092")
        buf.write("\3\2\2\2\u0096\u0099\3\2\2\2\u0097\u0095\3\2\2\2\u0097")
        buf.write("\u0098\3\2\2\2\u0098\27\3\2\2\2\u0099\u0097\3\2\2\2\u009a")
        buf.write("\u009c\7\4\2\2\u009b\u009d\5\32\16\2\u009c\u009b\3\2\2")
        buf.write("\2\u009c\u009d\3\2\2\2\u009d\u009e\3\2\2\2\u009e\u00a1")
        buf.write("\7\5\2\2\u009f\u00a1\3\2\2\2\u00a0\u009a\3\2\2\2\u00a0")
        buf.write("\u009f\3\2\2\2\u00a1\31\3\2\2\2\u00a2\u00a3\5\34\17\2")
        buf.write("\u00a3\u00a4\7\24\2\2\u00a4\u00a5\5\32\16\2\u00a5\u00a8")
        buf.write("\3\2\2\2\u00a6\u00a8\5\34\17\2\u00a7\u00a2\3\2\2\2\u00a7")
        buf.write("\u00a6\3\2\2\2\u00a8\33\3\2\2\2\u00a9\u00ab\t\2\2\2\u00aa")
        buf.write("\u00a9\3\2\2\2\u00ab\u00ac\3\2\2\2\u00ac\u00aa\3\2\2\2")
        buf.write("\u00ac\u00ad\3\2\2\2\u00ad\u00ae\3\2\2\2\u00ae\u00b0\7")
        buf.write("\23\2\2\u00af\u00b1\t\2\2\2\u00b0\u00af\3\2\2\2\u00b1")
        buf.write("\u00b2\3\2\2\2\u00b2\u00b0\3\2\2\2\u00b2\u00b3\3\2\2\2")
        buf.write("\u00b3\u00ba\3\2\2\2\u00b4\u00b6\t\2\2\2\u00b5\u00b4\3")
        buf.write("\2\2\2\u00b6\u00b7\3\2\2\2\u00b7\u00b5\3\2\2\2\u00b7\u00b8")
        buf.write("\3\2\2\2\u00b8\u00ba\3\2\2\2\u00b9\u00aa\3\2\2\2\u00b9")
        buf.write("\u00b5\3\2\2\2\u00ba\35\3\2\2\2\u00bb\u00bc\7\21\2\2\u00bc")
        buf.write("\u00bd\t\2\2\2\u00bd\u00be\t\3\2\2\u00be\u00bf\t\2\2\2")
        buf.write("\u00bf\u00c6\7\22\2\2\u00c0\u00c1\7\21\2\2\u00c1\u00c2")
        buf.write("\t\2\2\2\u00c2\u00c3\7\25\2\2\u00c3\u00c4\t\2\2\2\u00c4")
        buf.write("\u00c6\7\22\2\2\u00c5\u00bb\3\2\2\2\u00c5\u00c0\3\2\2")
        buf.write("\2\u00c6\37\3\2\2\2\u00c7\u00c8\7\30\2\2\u00c8!\3\2\2")
        buf.write("\2\27\'\65Q\\bfjw{\177\u0083\u0090\u0097\u009c\u00a0\u00a7")
        buf.write("\u00ac\u00b2\u00b7\u00b9\u00c5")
        return buf.getvalue()


class TikzParser ( Parser ):

    grammarFileName = "Tikz.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "'--'", "'['", "']'", "'/.'", "'style'", 
                     "'{'", "'}'", "'\\begin{tikzpicture}'", "'\\end{tikzpicture}'", 
                     "'\\node'", "'\\draw'", "'at'", "'and'", "'every'", 
                     "'('", "')'", "'='", "','", "':'", "';'", "'\\pause'" ]

    symbolicNames = [ "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "BEGINTIKZPICTURE", "ENDTIKZPICTURE", "NODE", "DRAW", 
                      "AT", "AND", "EVERY", "OPEN_PARANTHESES", "CLOSE_PARANTHESES", 
                      "EQUAL_TO", "COMMA", "COLON", "SEMICOLON", "PAUSE", 
                      "LABEL_VARIABLE", "EXPRESSION", "VARIABLE", "COMMENT", 
                      "WS" ]

    RULE_begin = 0
    RULE_instructions = 1
    RULE_draw = 2
    RULE_radius = 3
    RULE_nodeList = 4
    RULE_edgeNode = 5
    RULE_edgeProperties = 6
    RULE_node = 7
    RULE_nodeId = 8
    RULE_allGlobalProperties = 9
    RULE_globalProperties = 10
    RULE_nodeProperties = 11
    RULE_properties = 12
    RULE_individualProperty = 13
    RULE_coordinates = 14
    RULE_label = 15

    ruleNames =  [ "begin", "instructions", "draw", "radius", "nodeList", 
                   "edgeNode", "edgeProperties", "node", "nodeId", "allGlobalProperties", 
                   "globalProperties", "nodeProperties", "properties", "individualProperty", 
                   "coordinates", "label" ]

    EOF = Token.EOF
    T__0=1
    T__1=2
    T__2=3
    T__3=4
    T__4=5
    T__5=6
    T__6=7
    BEGINTIKZPICTURE=8
    ENDTIKZPICTURE=9
    NODE=10
    DRAW=11
    AT=12
    AND=13
    EVERY=14
    OPEN_PARANTHESES=15
    CLOSE_PARANTHESES=16
    EQUAL_TO=17
    COMMA=18
    COLON=19
    SEMICOLON=20
    PAUSE=21
    LABEL_VARIABLE=22
    EXPRESSION=23
    VARIABLE=24
    COMMENT=25
    WS=26

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.7.2")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None




    class BeginContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def BEGINTIKZPICTURE(self):
            return self.getToken(TikzParser.BEGINTIKZPICTURE, 0)

        def allGlobalProperties(self):
            return self.getTypedRuleContext(TikzParser.AllGlobalPropertiesContext,0)


        def ENDTIKZPICTURE(self):
            return self.getToken(TikzParser.ENDTIKZPICTURE, 0)

        def EOF(self):
            return self.getToken(TikzParser.EOF, 0)

        def instructions(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(TikzParser.InstructionsContext)
            else:
                return self.getTypedRuleContext(TikzParser.InstructionsContext,i)


        def getRuleIndex(self):
            return TikzParser.RULE_begin

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterBegin" ):
                listener.enterBegin(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitBegin" ):
                listener.exitBegin(self)




    def begin(self):

        localctx = TikzParser.BeginContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_begin)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 32
            self.match(TikzParser.BEGINTIKZPICTURE)
            self.state = 33
            self.allGlobalProperties()
            self.state = 37
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==TikzParser.NODE or _la==TikzParser.DRAW:
                self.state = 34
                self.instructions()
                self.state = 39
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 40
            self.match(TikzParser.ENDTIKZPICTURE)
            self.state = 41
            self.match(TikzParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class InstructionsContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def node(self):
            return self.getTypedRuleContext(TikzParser.NodeContext,0)


        def instructions(self):
            return self.getTypedRuleContext(TikzParser.InstructionsContext,0)


        def draw(self):
            return self.getTypedRuleContext(TikzParser.DrawContext,0)


        def getRuleIndex(self):
            return TikzParser.RULE_instructions

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterInstructions" ):
                listener.enterInstructions(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitInstructions" ):
                listener.exitInstructions(self)




    def instructions(self):

        localctx = TikzParser.InstructionsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_instructions)
        try:
            self.state = 51
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,1,self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 43
                self.node()
                self.state = 44
                self.instructions()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 46
                self.draw()
                self.state = 47
                self.instructions()
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 49
                self.draw()
                pass

            elif la_ == 4:
                self.enterOuterAlt(localctx, 4)
                self.state = 50
                self.node()
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class DrawContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def DRAW(self):
            return self.getToken(TikzParser.DRAW, 0)

        def edgeProperties(self):
            return self.getTypedRuleContext(TikzParser.EdgePropertiesContext,0)


        def nodeList(self):
            return self.getTypedRuleContext(TikzParser.NodeListContext,0)


        def SEMICOLON(self):
            return self.getToken(TikzParser.SEMICOLON, 0)

        def coordinates(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(TikzParser.CoordinatesContext)
            else:
                return self.getTypedRuleContext(TikzParser.CoordinatesContext,i)


        def VARIABLE(self):
            return self.getToken(TikzParser.VARIABLE, 0)

        def EXPRESSION(self):
            return self.getToken(TikzParser.EXPRESSION, 0)

        def radius(self):
            return self.getTypedRuleContext(TikzParser.RadiusContext,0)


        def nodeProperties(self):
            return self.getTypedRuleContext(TikzParser.NodePropertiesContext,0)


        def label(self):
            return self.getTypedRuleContext(TikzParser.LabelContext,0)


        def getRuleIndex(self):
            return TikzParser.RULE_draw

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterDraw" ):
                listener.enterDraw(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitDraw" ):
                listener.exitDraw(self)




    def draw(self):

        localctx = TikzParser.DrawContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_draw)
        self._la = 0 # Token type
        try:
            self.state = 79
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,2,self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 53
                self.match(TikzParser.DRAW)
                self.state = 54
                self.edgeProperties()
                self.state = 55
                self.nodeList()
                self.state = 56
                self.match(TikzParser.SEMICOLON)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 58
                self.match(TikzParser.DRAW)
                self.state = 59
                self.edgeProperties()
                self.state = 60
                self.coordinates()
                self.state = 61
                _la = self._input.LA(1)
                if not(_la==TikzParser.EXPRESSION or _la==TikzParser.VARIABLE):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                self.state = 62
                self.coordinates()
                self.state = 63
                self.match(TikzParser.SEMICOLON)
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 65
                self.match(TikzParser.DRAW)
                self.state = 66
                self.edgeProperties()
                self.state = 67
                self.coordinates()
                self.state = 68
                _la = self._input.LA(1)
                if not(_la==TikzParser.EXPRESSION or _la==TikzParser.VARIABLE):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                self.state = 69
                self.radius()
                self.state = 70
                self.match(TikzParser.SEMICOLON)
                pass

            elif la_ == 4:
                self.enterOuterAlt(localctx, 4)
                self.state = 72
                self.match(TikzParser.DRAW)
                self.state = 73
                self.edgeProperties()
                self.state = 74
                self.coordinates()
                self.state = 75
                _la = self._input.LA(1)
                if not(_la==TikzParser.EXPRESSION or _la==TikzParser.VARIABLE):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                self.state = 76
                self.nodeProperties()
                self.state = 77
                self.label()
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class RadiusContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def OPEN_PARANTHESES(self):
            return self.getToken(TikzParser.OPEN_PARANTHESES, 0)

        def CLOSE_PARANTHESES(self):
            return self.getToken(TikzParser.CLOSE_PARANTHESES, 0)

        def VARIABLE(self):
            return self.getToken(TikzParser.VARIABLE, 0)

        def EXPRESSION(self):
            return self.getToken(TikzParser.EXPRESSION, 0)

        def getRuleIndex(self):
            return TikzParser.RULE_radius

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterRadius" ):
                listener.enterRadius(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitRadius" ):
                listener.exitRadius(self)




    def radius(self):

        localctx = TikzParser.RadiusContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_radius)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 81
            self.match(TikzParser.OPEN_PARANTHESES)
            self.state = 82
            _la = self._input.LA(1)
            if not(_la==TikzParser.EXPRESSION or _la==TikzParser.VARIABLE):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
            self.state = 83
            self.match(TikzParser.CLOSE_PARANTHESES)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class NodeListContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def edgeNode(self):
            return self.getTypedRuleContext(TikzParser.EdgeNodeContext,0)


        def nodeList(self):
            return self.getTypedRuleContext(TikzParser.NodeListContext,0)


        def getRuleIndex(self):
            return TikzParser.RULE_nodeList

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterNodeList" ):
                listener.enterNodeList(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitNodeList" ):
                listener.exitNodeList(self)




    def nodeList(self):

        localctx = TikzParser.NodeListContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_nodeList)
        try:
            self.state = 90
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,3,self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 85
                self.edgeNode()
                self.state = 86
                self.match(TikzParser.T__0)
                self.state = 87
                self.nodeList()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 89
                self.edgeNode()
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class EdgeNodeContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def coordinates(self):
            return self.getTypedRuleContext(TikzParser.CoordinatesContext,0)


        def OPEN_PARANTHESES(self):
            return self.getToken(TikzParser.OPEN_PARANTHESES, 0)

        def CLOSE_PARANTHESES(self):
            return self.getToken(TikzParser.CLOSE_PARANTHESES, 0)

        def VARIABLE(self):
            return self.getToken(TikzParser.VARIABLE, 0)

        def EXPRESSION(self):
            return self.getToken(TikzParser.EXPRESSION, 0)

        def getRuleIndex(self):
            return TikzParser.RULE_edgeNode

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterEdgeNode" ):
                listener.enterEdgeNode(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitEdgeNode" ):
                listener.exitEdgeNode(self)




    def edgeNode(self):

        localctx = TikzParser.EdgeNodeContext(self, self._ctx, self.state)
        self.enterRule(localctx, 10, self.RULE_edgeNode)
        self._la = 0 # Token type
        try:
            self.state = 96
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,4,self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 92
                self.coordinates()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 93
                self.match(TikzParser.OPEN_PARANTHESES)
                self.state = 94
                _la = self._input.LA(1)
                if not(_la==TikzParser.EXPRESSION or _la==TikzParser.VARIABLE):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                self.state = 95
                self.match(TikzParser.CLOSE_PARANTHESES)
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class EdgePropertiesContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def properties(self):
            return self.getTypedRuleContext(TikzParser.PropertiesContext,0)


        def getRuleIndex(self):
            return TikzParser.RULE_edgeProperties

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterEdgeProperties" ):
                listener.enterEdgeProperties(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitEdgeProperties" ):
                listener.exitEdgeProperties(self)




    def edgeProperties(self):

        localctx = TikzParser.EdgePropertiesContext(self, self._ctx, self.state)
        self.enterRule(localctx, 12, self.RULE_edgeProperties)
        self._la = 0 # Token type
        try:
            self.state = 104
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [TikzParser.T__1]:
                self.enterOuterAlt(localctx, 1)
                self.state = 98
                self.match(TikzParser.T__1)
                self.state = 100
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==TikzParser.EXPRESSION or _la==TikzParser.VARIABLE:
                    self.state = 99
                    self.properties()


                self.state = 102
                self.match(TikzParser.T__2)
                pass
            elif token in [TikzParser.OPEN_PARANTHESES]:
                self.enterOuterAlt(localctx, 2)

                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class NodeContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NODE(self):
            return self.getToken(TikzParser.NODE, 0)

        def nodeProperties(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(TikzParser.NodePropertiesContext)
            else:
                return self.getTypedRuleContext(TikzParser.NodePropertiesContext,i)


        def nodeId(self):
            return self.getTypedRuleContext(TikzParser.NodeIdContext,0)


        def AT(self):
            return self.getToken(TikzParser.AT, 0)

        def coordinates(self):
            return self.getTypedRuleContext(TikzParser.CoordinatesContext,0)


        def label(self):
            return self.getTypedRuleContext(TikzParser.LabelContext,0)


        def getRuleIndex(self):
            return TikzParser.RULE_node

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterNode" ):
                listener.enterNode(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitNode" ):
                listener.exitNode(self)




    def node(self):

        localctx = TikzParser.NodeContext(self, self._ctx, self.state)
        self.enterRule(localctx, 14, self.RULE_node)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 106
            self.match(TikzParser.NODE)
            self.state = 107
            self.nodeProperties()
            self.state = 108
            self.nodeId()
            self.state = 109
            self.nodeProperties()
            self.state = 110
            self.match(TikzParser.AT)
            self.state = 111
            self.coordinates()
            self.state = 112
            self.nodeProperties()
            self.state = 113
            self.label()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class NodeIdContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def OPEN_PARANTHESES(self):
            return self.getToken(TikzParser.OPEN_PARANTHESES, 0)

        def CLOSE_PARANTHESES(self):
            return self.getToken(TikzParser.CLOSE_PARANTHESES, 0)

        def VARIABLE(self):
            return self.getToken(TikzParser.VARIABLE, 0)

        def EXPRESSION(self):
            return self.getToken(TikzParser.EXPRESSION, 0)

        def getRuleIndex(self):
            return TikzParser.RULE_nodeId

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterNodeId" ):
                listener.enterNodeId(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitNodeId" ):
                listener.exitNodeId(self)




    def nodeId(self):

        localctx = TikzParser.NodeIdContext(self, self._ctx, self.state)
        self.enterRule(localctx, 16, self.RULE_nodeId)
        self._la = 0 # Token type
        try:
            self.state = 121
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [TikzParser.OPEN_PARANTHESES]:
                self.enterOuterAlt(localctx, 1)
                self.state = 115
                self.match(TikzParser.OPEN_PARANTHESES)
                self.state = 117
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==TikzParser.EXPRESSION or _la==TikzParser.VARIABLE:
                    self.state = 116
                    _la = self._input.LA(1)
                    if not(_la==TikzParser.EXPRESSION or _la==TikzParser.VARIABLE):
                        self._errHandler.recoverInline(self)
                    else:
                        self._errHandler.reportMatch(self)
                        self.consume()


                self.state = 119
                self.match(TikzParser.CLOSE_PARANTHESES)
                pass
            elif token in [TikzParser.T__1, TikzParser.AT]:
                self.enterOuterAlt(localctx, 2)

                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class AllGlobalPropertiesContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def globalProperties(self):
            return self.getTypedRuleContext(TikzParser.GlobalPropertiesContext,0)


        def getRuleIndex(self):
            return TikzParser.RULE_allGlobalProperties

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterAllGlobalProperties" ):
                listener.enterAllGlobalProperties(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitAllGlobalProperties" ):
                listener.exitAllGlobalProperties(self)




    def allGlobalProperties(self):

        localctx = TikzParser.AllGlobalPropertiesContext(self, self._ctx, self.state)
        self.enterRule(localctx, 18, self.RULE_allGlobalProperties)
        self._la = 0 # Token type
        try:
            self.state = 129
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [TikzParser.T__1]:
                self.enterOuterAlt(localctx, 1)
                self.state = 123
                self.match(TikzParser.T__1)
                self.state = 125
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << TikzParser.EVERY) | (1 << TikzParser.EXPRESSION) | (1 << TikzParser.VARIABLE))) != 0):
                    self.state = 124
                    self.globalProperties(0)


                self.state = 127
                self.match(TikzParser.T__2)
                pass
            elif token in [TikzParser.ENDTIKZPICTURE, TikzParser.NODE, TikzParser.DRAW]:
                self.enterOuterAlt(localctx, 2)

                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class GlobalPropertiesContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def EVERY(self):
            return self.getToken(TikzParser.EVERY, 0)

        def EQUAL_TO(self):
            return self.getToken(TikzParser.EQUAL_TO, 0)

        def properties(self):
            return self.getTypedRuleContext(TikzParser.PropertiesContext,0)


        def VARIABLE(self):
            return self.getToken(TikzParser.VARIABLE, 0)

        def EXPRESSION(self):
            return self.getToken(TikzParser.EXPRESSION, 0)

        def globalProperties(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(TikzParser.GlobalPropertiesContext)
            else:
                return self.getTypedRuleContext(TikzParser.GlobalPropertiesContext,i)


        def COMMA(self):
            return self.getToken(TikzParser.COMMA, 0)

        def getRuleIndex(self):
            return TikzParser.RULE_globalProperties

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterGlobalProperties" ):
                listener.enterGlobalProperties(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitGlobalProperties" ):
                listener.exitGlobalProperties(self)



    def globalProperties(self, _p:int=0):
        _parentctx = self._ctx
        _parentState = self.state
        localctx = TikzParser.GlobalPropertiesContext(self, self._ctx, _parentState)
        _prevctx = localctx
        _startState = 20
        self.enterRecursionRule(localctx, 20, self.RULE_globalProperties, _p)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 142
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [TikzParser.EVERY]:
                self.state = 132
                self.match(TikzParser.EVERY)
                self.state = 133
                _la = self._input.LA(1)
                if not(_la==TikzParser.EXPRESSION or _la==TikzParser.VARIABLE):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                self.state = 134
                self.match(TikzParser.T__3)
                self.state = 135
                self.match(TikzParser.T__4)
                self.state = 136
                self.match(TikzParser.EQUAL_TO)
                self.state = 137
                self.match(TikzParser.T__5)
                self.state = 138
                self.properties()
                self.state = 139
                self.match(TikzParser.T__6)
                pass
            elif token in [TikzParser.EXPRESSION, TikzParser.VARIABLE]:
                self.state = 141
                self.properties()
                pass
            else:
                raise NoViableAltException(self)

            self._ctx.stop = self._input.LT(-1)
            self.state = 149
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,12,self._ctx)
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1:
                    if self._parseListeners is not None:
                        self.triggerExitRuleEvent()
                    _prevctx = localctx
                    localctx = TikzParser.GlobalPropertiesContext(self, _parentctx, _parentState)
                    self.pushNewRecursionContext(localctx, _startState, self.RULE_globalProperties)
                    self.state = 144
                    if not self.precpred(self._ctx, 3):
                        from antlr4.error.Errors import FailedPredicateException
                        raise FailedPredicateException(self, "self.precpred(self._ctx, 3)")
                    self.state = 145
                    self.match(TikzParser.COMMA)
                    self.state = 146
                    self.globalProperties(4) 
                self.state = 151
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,12,self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.unrollRecursionContexts(_parentctx)
        return localctx


    class NodePropertiesContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def properties(self):
            return self.getTypedRuleContext(TikzParser.PropertiesContext,0)


        def getRuleIndex(self):
            return TikzParser.RULE_nodeProperties

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterNodeProperties" ):
                listener.enterNodeProperties(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitNodeProperties" ):
                listener.exitNodeProperties(self)




    def nodeProperties(self):

        localctx = TikzParser.NodePropertiesContext(self, self._ctx, self.state)
        self.enterRule(localctx, 22, self.RULE_nodeProperties)
        self._la = 0 # Token type
        try:
            self.state = 158
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,14,self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 152
                self.match(TikzParser.T__1)
                self.state = 154
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==TikzParser.EXPRESSION or _la==TikzParser.VARIABLE:
                    self.state = 153
                    self.properties()


                self.state = 156
                self.match(TikzParser.T__2)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)

                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class PropertiesContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def individualProperty(self):
            return self.getTypedRuleContext(TikzParser.IndividualPropertyContext,0)


        def COMMA(self):
            return self.getToken(TikzParser.COMMA, 0)

        def properties(self):
            return self.getTypedRuleContext(TikzParser.PropertiesContext,0)


        def getRuleIndex(self):
            return TikzParser.RULE_properties

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterProperties" ):
                listener.enterProperties(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitProperties" ):
                listener.exitProperties(self)




    def properties(self):

        localctx = TikzParser.PropertiesContext(self, self._ctx, self.state)
        self.enterRule(localctx, 24, self.RULE_properties)
        try:
            self.state = 165
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,15,self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 160
                self.individualProperty()
                self.state = 161
                self.match(TikzParser.COMMA)
                self.state = 162
                self.properties()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 164
                self.individualProperty()
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class IndividualPropertyContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def EQUAL_TO(self):
            return self.getToken(TikzParser.EQUAL_TO, 0)

        def VARIABLE(self, i:int=None):
            if i is None:
                return self.getTokens(TikzParser.VARIABLE)
            else:
                return self.getToken(TikzParser.VARIABLE, i)

        def EXPRESSION(self, i:int=None):
            if i is None:
                return self.getTokens(TikzParser.EXPRESSION)
            else:
                return self.getToken(TikzParser.EXPRESSION, i)

        def getRuleIndex(self):
            return TikzParser.RULE_individualProperty

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterIndividualProperty" ):
                listener.enterIndividualProperty(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitIndividualProperty" ):
                listener.exitIndividualProperty(self)




    def individualProperty(self):

        localctx = TikzParser.IndividualPropertyContext(self, self._ctx, self.state)
        self.enterRule(localctx, 26, self.RULE_individualProperty)
        self._la = 0 # Token type
        try:
            self.state = 183
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,19,self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 168 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while True:
                    self.state = 167
                    _la = self._input.LA(1)
                    if not(_la==TikzParser.EXPRESSION or _la==TikzParser.VARIABLE):
                        self._errHandler.recoverInline(self)
                    else:
                        self._errHandler.reportMatch(self)
                        self.consume()
                    self.state = 170 
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)
                    if not (_la==TikzParser.EXPRESSION or _la==TikzParser.VARIABLE):
                        break

                self.state = 172
                self.match(TikzParser.EQUAL_TO)
                self.state = 174 
                self._errHandler.sync(self)
                _alt = 1
                while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                    if _alt == 1:
                        self.state = 173
                        _la = self._input.LA(1)
                        if not(_la==TikzParser.EXPRESSION or _la==TikzParser.VARIABLE):
                            self._errHandler.recoverInline(self)
                        else:
                            self._errHandler.reportMatch(self)
                            self.consume()

                    else:
                        raise NoViableAltException(self)
                    self.state = 176 
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input,17,self._ctx)

                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 179 
                self._errHandler.sync(self)
                _alt = 1
                while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                    if _alt == 1:
                        self.state = 178
                        _la = self._input.LA(1)
                        if not(_la==TikzParser.EXPRESSION or _la==TikzParser.VARIABLE):
                            self._errHandler.recoverInline(self)
                        else:
                            self._errHandler.reportMatch(self)
                            self.consume()

                    else:
                        raise NoViableAltException(self)
                    self.state = 181 
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input,18,self._ctx)

                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class CoordinatesContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return TikzParser.RULE_coordinates

     
        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)



    class CartesianCoordinatesContext(CoordinatesContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a TikzParser.CoordinatesContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def OPEN_PARANTHESES(self):
            return self.getToken(TikzParser.OPEN_PARANTHESES, 0)
        def CLOSE_PARANTHESES(self):
            return self.getToken(TikzParser.CLOSE_PARANTHESES, 0)
        def VARIABLE(self, i:int=None):
            if i is None:
                return self.getTokens(TikzParser.VARIABLE)
            else:
                return self.getToken(TikzParser.VARIABLE, i)
        def EXPRESSION(self, i:int=None):
            if i is None:
                return self.getTokens(TikzParser.EXPRESSION)
            else:
                return self.getToken(TikzParser.EXPRESSION, i)
        def COMMA(self):
            return self.getToken(TikzParser.COMMA, 0)
        def AND(self):
            return self.getToken(TikzParser.AND, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterCartesianCoordinates" ):
                listener.enterCartesianCoordinates(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitCartesianCoordinates" ):
                listener.exitCartesianCoordinates(self)


    class PolarCoordinatesContext(CoordinatesContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a TikzParser.CoordinatesContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def OPEN_PARANTHESES(self):
            return self.getToken(TikzParser.OPEN_PARANTHESES, 0)
        def COLON(self):
            return self.getToken(TikzParser.COLON, 0)
        def CLOSE_PARANTHESES(self):
            return self.getToken(TikzParser.CLOSE_PARANTHESES, 0)
        def VARIABLE(self, i:int=None):
            if i is None:
                return self.getTokens(TikzParser.VARIABLE)
            else:
                return self.getToken(TikzParser.VARIABLE, i)
        def EXPRESSION(self, i:int=None):
            if i is None:
                return self.getTokens(TikzParser.EXPRESSION)
            else:
                return self.getToken(TikzParser.EXPRESSION, i)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPolarCoordinates" ):
                listener.enterPolarCoordinates(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPolarCoordinates" ):
                listener.exitPolarCoordinates(self)



    def coordinates(self):

        localctx = TikzParser.CoordinatesContext(self, self._ctx, self.state)
        self.enterRule(localctx, 28, self.RULE_coordinates)
        self._la = 0 # Token type
        try:
            self.state = 195
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,20,self._ctx)
            if la_ == 1:
                localctx = TikzParser.CartesianCoordinatesContext(self, localctx)
                self.enterOuterAlt(localctx, 1)
                self.state = 185
                self.match(TikzParser.OPEN_PARANTHESES)
                self.state = 186
                _la = self._input.LA(1)
                if not(_la==TikzParser.EXPRESSION or _la==TikzParser.VARIABLE):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                self.state = 187
                _la = self._input.LA(1)
                if not(_la==TikzParser.AND or _la==TikzParser.COMMA):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                self.state = 188
                _la = self._input.LA(1)
                if not(_la==TikzParser.EXPRESSION or _la==TikzParser.VARIABLE):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                self.state = 189
                self.match(TikzParser.CLOSE_PARANTHESES)
                pass

            elif la_ == 2:
                localctx = TikzParser.PolarCoordinatesContext(self, localctx)
                self.enterOuterAlt(localctx, 2)
                self.state = 190
                self.match(TikzParser.OPEN_PARANTHESES)
                self.state = 191
                _la = self._input.LA(1)
                if not(_la==TikzParser.EXPRESSION or _la==TikzParser.VARIABLE):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                self.state = 192
                self.match(TikzParser.COLON)
                self.state = 193
                _la = self._input.LA(1)
                if not(_la==TikzParser.EXPRESSION or _la==TikzParser.VARIABLE):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                self.state = 194
                self.match(TikzParser.CLOSE_PARANTHESES)
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class LabelContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def LABEL_VARIABLE(self):
            return self.getToken(TikzParser.LABEL_VARIABLE, 0)

        def getRuleIndex(self):
            return TikzParser.RULE_label

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterLabel" ):
                listener.enterLabel(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitLabel" ):
                listener.exitLabel(self)




    def label(self):

        localctx = TikzParser.LabelContext(self, self._ctx, self.state)
        self.enterRule(localctx, 30, self.RULE_label)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 197
            self.match(TikzParser.LABEL_VARIABLE)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx



    def sempred(self, localctx:RuleContext, ruleIndex:int, predIndex:int):
        if self._predicates == None:
            self._predicates = dict()
        self._predicates[10] = self.globalProperties_sempred
        pred = self._predicates.get(ruleIndex, None)
        if pred is None:
            raise Exception("No predicate with index:" + str(ruleIndex))
        else:
            return pred(localctx, predIndex)

    def globalProperties_sempred(self, localctx:GlobalPropertiesContext, predIndex:int):
            if predIndex == 0:
                return self.precpred(self._ctx, 3)
         







# Generated from tikz2graphml/grammar/Tikz.g4 by ANTLR 4.7.2
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .TikzParser import TikzParser
else:
    from TikzParser import TikzParser

# This class defines a complete listener for a parse tree produced by TikzParser.
class TikzListener(ParseTreeListener):

    # Enter a parse tree produced by TikzParser#begin.
    def enterBegin(self, ctx:TikzParser.BeginContext):
        pass

    # Exit a parse tree produced by TikzParser#begin.
    def exitBegin(self, ctx:TikzParser.BeginContext):
        pass


    # Enter a parse tree produced by TikzParser#instructions.
    def enterInstructions(self, ctx:TikzParser.InstructionsContext):
        pass

    # Exit a parse tree produced by TikzParser#instructions.
    def exitInstructions(self, ctx:TikzParser.InstructionsContext):
        pass


    # Enter a parse tree produced by TikzParser#draw.
    def enterDraw(self, ctx:TikzParser.DrawContext):
        pass

    # Exit a parse tree produced by TikzParser#draw.
    def exitDraw(self, ctx:TikzParser.DrawContext):
        pass


    # Enter a parse tree produced by TikzParser#radius.
    def enterRadius(self, ctx:TikzParser.RadiusContext):
        pass

    # Exit a parse tree produced by TikzParser#radius.
    def exitRadius(self, ctx:TikzParser.RadiusContext):
        pass


    # Enter a parse tree produced by TikzParser#nodeList.
    def enterNodeList(self, ctx:TikzParser.NodeListContext):
        pass

    # Exit a parse tree produced by TikzParser#nodeList.
    def exitNodeList(self, ctx:TikzParser.NodeListContext):
        pass


    # Enter a parse tree produced by TikzParser#edgeNode.
    def enterEdgeNode(self, ctx:TikzParser.EdgeNodeContext):
        pass

    # Exit a parse tree produced by TikzParser#edgeNode.
    def exitEdgeNode(self, ctx:TikzParser.EdgeNodeContext):
        pass


    # Enter a parse tree produced by TikzParser#edgeProperties.
    def enterEdgeProperties(self, ctx:TikzParser.EdgePropertiesContext):
        pass

    # Exit a parse tree produced by TikzParser#edgeProperties.
    def exitEdgeProperties(self, ctx:TikzParser.EdgePropertiesContext):
        pass


    # Enter a parse tree produced by TikzParser#node.
    def enterNode(self, ctx:TikzParser.NodeContext):
        pass

    # Exit a parse tree produced by TikzParser#node.
    def exitNode(self, ctx:TikzParser.NodeContext):
        pass


    # Enter a parse tree produced by TikzParser#nodeId.
    def enterNodeId(self, ctx:TikzParser.NodeIdContext):
        pass

    # Exit a parse tree produced by TikzParser#nodeId.
    def exitNodeId(self, ctx:TikzParser.NodeIdContext):
        pass


    # Enter a parse tree produced by TikzParser#allGlobalProperties.
    def enterAllGlobalProperties(self, ctx:TikzParser.AllGlobalPropertiesContext):
        pass

    # Exit a parse tree produced by TikzParser#allGlobalProperties.
    def exitAllGlobalProperties(self, ctx:TikzParser.AllGlobalPropertiesContext):
        pass


    # Enter a parse tree produced by TikzParser#globalProperties.
    def enterGlobalProperties(self, ctx:TikzParser.GlobalPropertiesContext):
        pass

    # Exit a parse tree produced by TikzParser#globalProperties.
    def exitGlobalProperties(self, ctx:TikzParser.GlobalPropertiesContext):
        pass


    # Enter a parse tree produced by TikzParser#nodeProperties.
    def enterNodeProperties(self, ctx:TikzParser.NodePropertiesContext):
        pass

    # Exit a parse tree produced by TikzParser#nodeProperties.
    def exitNodeProperties(self, ctx:TikzParser.NodePropertiesContext):
        pass


    # Enter a parse tree produced by TikzParser#properties.
    def enterProperties(self, ctx:TikzParser.PropertiesContext):
        pass

    # Exit a parse tree produced by TikzParser#properties.
    def exitProperties(self, ctx:TikzParser.PropertiesContext):
        pass


    # Enter a parse tree produced by TikzParser#individualProperty.
    def enterIndividualProperty(self, ctx:TikzParser.IndividualPropertyContext):
        pass

    # Exit a parse tree produced by TikzParser#individualProperty.
    def exitIndividualProperty(self, ctx:TikzParser.IndividualPropertyContext):
        pass


    # Enter a parse tree produced by TikzParser#cartesianCoordinates.
    def enterCartesianCoordinates(self, ctx:TikzParser.CartesianCoordinatesContext):
        pass

    # Exit a parse tree produced by TikzParser#cartesianCoordinates.
    def exitCartesianCoordinates(self, ctx:TikzParser.CartesianCoordinatesContext):
        pass


    # Enter a parse tree produced by TikzParser#polarCoordinates.
    def enterPolarCoordinates(self, ctx:TikzParser.PolarCoordinatesContext):
        pass

    # Exit a parse tree produced by TikzParser#polarCoordinates.
    def exitPolarCoordinates(self, ctx:TikzParser.PolarCoordinatesContext):
        pass


    # Enter a parse tree produced by TikzParser#label.
    def enterLabel(self, ctx:TikzParser.LabelContext):
        pass

    # Exit a parse tree produced by TikzParser#label.
    def exitLabel(self, ctx:TikzParser.LabelContext):
        pass





import re
import copy
import logging

logger = logging.getLogger()

# This function is range function for floating points
# Needed because "range(0.5, 10.5, 0.5)" : Not supported by default range function
def floatRange(start, end, step):
    # Note: end is inclusive
    while(start <= end):
        yield start
        start += step

# Unrolls the loop completely
# Replace variables with their values
# Example:
# for i in {1,2,3,4} { A[i] = i; } 
# will be replaced by
#    A[1] = i;  
#    A[2] = i;  
#    A[3] = i;  
#    A[4] = i;  
def replaceVarsinForeach(foreachHead, block):
    block = block.strip().strip("{").strip("}")
    unrolledBlocks = ""
    # greps variable and their values in foreach statement
    #
    # Example 
    #
    # \foreach \from/\to in {B/t4,B/t5,C/t6,C/t7} { block }
    # Variables = {\from, \to}
    # Values {
    #    \from : B, B, C, C
    #    \to : t4, t5, t6, t7
    # }
    x = re.findall("\\\\foreach(.*)in[\s]*{(.*?)}", foreachHead)
    if(x):
        x = x[0]
        if(x[1].__contains__("...")):
            variable = x[0].strip()
            varRange = x[1].strip()
            rangeOfVariable = varRange.split(",")
            start = float(rangeOfVariable[0])
            if(isinstance(rangeOfVariable[1], float)):
                # 1,2.....10
                step = float(rangeOfVariable[1])-float(rangeOfVariable[0])
            else:
                # 1.....10
                step = 1
            end = float(rangeOfVariable[-1])
            for val in floatRange(start, end, step): 
                remainblock = copy.copy(block)
                remainblock = re.sub("\\"+variable.strip(), str(val), remainblock)
                unrolledBlocks += remainblock.strip() + "\n"
        else:
            varList = x[0].split("/")
            var_index_to_name = {}
            rangeOfVariables = x[1].strip().split(",")
            for i, var in enumerate(varList):
                var_index_to_name[i] = var.strip().strip("\\")
            # \x/\y in 1/2,3/4,5/6,7/8
            # its like x and y pairwise (1,2) then (3,4) then so on
            for rangeValues in rangeOfVariables:
                remainblock = copy.copy(block)
                for i, val in enumerate(rangeValues.strip().split("/")):
                    # Subsitute variable with its value in entire block
                    remainblock = re.sub("\\\\"+var_index_to_name[i], val, remainblock)
                unrolledBlocks += remainblock.strip() + "\n"
    return unrolledBlocks

# Find Foreach block in code and then copy blocks and in each block replace variables by their values
def parseAndHandleForEach(tikzBlock):
    # Find all the foreach and its corresponding block in tikz code and call the replaceVars function
    # \foreach \from/\to in {B/t4,B/t5,C/t6,C/t7} 
    #     block
    # Not block may contain foreach statement
    regexForForeach = re.compile("\\\\foreach.*?in[\s]*{.*?}[\s]*", re.MULTILINE)
    firstForeach = None
    for x in re.finditer(regexForForeach, tikzBlock):
        firstForeach = x
        break
    if not firstForeach:
        return tikzBlock
    block = ""
    parantheses = 0
    startswithParantheses = False
    startOffset = 0
    if tikzBlock[firstForeach.end(0)] == '{':
        block += "{"
        parantheses = 1
        startswithParantheses = True
        startOffset = 1
    for char in tikzBlock[firstForeach.end(0)+startOffset:]:
        block += char
        if char == "{":
            parantheses += 1
        elif char == "}":
            parantheses -= 1
            if parantheses == 0 and startswithParantheses:
                break
        elif char == ";" and parantheses == 0:
            break
    return tikzBlock[:firstForeach.start(0)] + replaceVarsinForeach(tikzBlock[firstForeach.start(0): firstForeach.end(0)], block) + tikzBlock[firstForeach.end(0)+len(block):]

# returns list of blocks with unrolled for loops
# A block is the code inside tikz begin and end
def getCodeInsideTIKZAfterUnrolling(filename):
    tikzBlocks = []
    with open(filename) as inputFile:
        fileContent = inputFile.read()
    regextToGetTikzPictureCode = re.compile("\\\\end[\s]*{tikzpicture}", re.MULTILINE)
    i = -1
    for block in regextToGetTikzPictureCode.split(fileContent):
        if(block.__contains__("tikzpicture")):
            for codeInsideTikzPicture in (re.split("\\\\begin{tikzpicture}", block)):
                i += 1
                if i % 2 == 1:
                    oneforeachRemoved = parseAndHandleForEach(codeInsideTikzPicture)
                    while(oneforeachRemoved.count("\\foreach")>0):
                        oneforeachRemoved = parseAndHandleForEach(oneforeachRemoved)
                    unrolledForeachInTikzPart = oneforeachRemoved
                    tikzBlocks.append("\\begin{tikzpicture}" + unrolledForeachInTikzPart + "\\end{tikzpicture}")
    return tikzBlocks


import re
import math
import logging
import numpy as np
import networkx as nx
import xml.dom.minidom
from matplotlib import colors
import tikz2graphml.pyyed as pyyed
from pylatexenc.latex2text import LatexNodes2Text

logger = logging.getLogger()

"""
Main Graph class storing all the properties of the edges and nodes of the graph
for a single TikZ graph code at a time
"""
class Graph:

	def __init__(self, scalingFactor: float):
		self.scalingFactor = scalingFactor
		self.G = pyyed.Graph()
		self.numNodes = 0
		self.numEdges = 0
		self.maxScaleFactor = 1.0
		self.nodes = []
		self.edges = []
		self.globalProperties = {}
		self.default_fontSize = str(int(0.20 * self.scalingFactor))
		self.defaultNodeSide = "10"

	# Rotate coordinates by a given angle(in Degrees)
	def rotateCoordinates(self, coordinates, angle):
		cosA = round(math.cos(math.radians(float(angle))), 10)
		sinA = round(math.sin(math.radians(float(angle))), 10)
		for index, val in enumerate(coordinates):
			x, y = val[0], val[1]
			coordinates[index][0] =      x * cosA + y * sinA
			coordinates[index][1] =  1 * x * sinA - y * cosA

	# Rescale coordinates by a given scale
	def rescaleCoordinates(self, coordinates, scale):
		for index, val in enumerate(coordinates):
			coordinates[index] = [float(val[0]) * scale, float(val[1]) * scale]

	# Rescale a coordinate tuple by a given scale
	def rescaleCoordinate(self, X, Y, scale):
		coordinates = np.array([[X, Y]])
		return (nx.rescale_layout(coordinates, scale))

	# Obtain scaling factor  for a given min distance
	# The formula used here is emperically tested so that it
	# works for most tikz code
	def getScalingFactor(self, minDistance):
		return self.maxScaleFactor * (3 - 2 * minDistance) * self.scalingFactor

	# Given a color string name, output it's corresponding hex-code
	def getColor(self, fill):
		if fill is None or fill == "none":
			return None
		m = re.search('^\s*([a-zA-Z]+)(?:!(\d+))?\s*$', fill)
		if not m:
			clr = None 	# stands for no color / transparent
		else:
			if m.group(2) is not None:
				alpha = float(m.group(2))
				clr = colors.to_hex(colors.to_rgba(m.group(1), alpha=alpha/256.0), keep_alpha=True)
			elif m.group(1) is not None:
				clr = colors.to_hex(colors.to_rgba(m.group(1)), keep_alpha=False)
			else:
				clr = None
		return clr

	# Helper function to add a node with its properties into the graph object
	def addNode(self, nodeID:str = None, X:str = "0", Y:str = "0", label:str = None,
		height:str = "-", width:str = "-", inner_sep:str = "", fill:str = "1", edge_color:str = None,
		scale:str = "1", shape:str = "rectangle", regular_polygon_sides:str="0", rotate:str="0", auto:str="center", transparent:bool="false"):
		if inner_sep == "":
			inner_sep = self.defaultNodeSide

		# This rotate is rotation of shape of node
		if rotate != "0":
			logger.warning("Rotation of each node cannot be done in graphml")

		fillClr = self.getColor(fill)
		edgeClr = self.getColor(edge_color)

		# When nodeID is unassigned for a given node, then the ID is generated
		# internally by incrementing the numNodes counter variable
		if nodeID is None:
			nodeID = str(self.numNodes)

		self.numNodes += 1

		# Given a label string, we convert it to text format
		# to be used in GraphML generation
		if label is not None:
			label = LatexNodes2Text().latex_to_text(label)

        # Currently the distance values for cm & pt  is handled together
		# and those without any units are handled separately
		if inner_sep.__contains__("cm"):
			m = re.search('^\s*([0-9/*-+.]+)\s*(?:pt|cm)?\s*$', inner_sep)
			if m and len(m.group(1)) > 0 and m.group(1) != ".":
				inner_sep = float(m.group(1))
		else:
			m = re.search('^\s*([0-9/*-+.]+)\s*(?:pt|cm)?\s*$', inner_sep)
			if m and len(m.group(1)) > 0 and m.group(1) != ".":
				inner_sep = max(float(m.group(1)), float(self.defaultNodeSide)) * 0.0352778

		if shape == "circle":
			shape = "ellipse"

		if shape is None:
			shape = "rectangle"

		# Creating a node dict for storing all its properties
		node = {
			"nodeID": nodeID,
			"shape": shape,
			"label": label,
			"X": float(X),
			"Y": float(Y),
			"shape_fill": fillClr,
			"edge_color": edgeClr,
			"height": height if height != '-' else inner_sep ,
			"width": width if width != '-' else inner_sep,
			"edge_width": "1.0",
			"transparent": transparent
		}
		logger.debug("Adding Node to Graph : \n{}".format(node))
		self.nodes.append(node)
		return nodeID

	# TODO: Add exception handling when NodeID is referenced without declaring it
	# Helper function to add a edge with its properties into the graph object
	def addEdge(self, nodeX:str=None, nodeY:str=None, arrowHead:bool=False, arrowFoot:bool=False, color:str="black", width:str="1", label:str="", line_type:str="line"):

		if line_type is not None and line_type == "solid":
			line_type="line"
		elif line_type is not None and line_type == "dash":
			line_type="dashed"

		# Creating a edge dict for storing all its properties
		self.edges.append({
			"x": nodeX,
			"y": nodeY,
			"arrowhead": "standard" if arrowHead==True else "none",
			"arrowfoot": "standard" if arrowFoot==True else "none",
			"color" : self.getColor(color),
			"width" : width,
			"label" : label,
			"line_type" : line_type
		})

	# Obtain a GraphML xml snippet for the existing graph object
	def get_graph(self):
		sz = len(self.nodes)
		if sz == 0:
			logger.warn("No Nodes are present in the graph. Printing Empty GraphML graph.")
			return ""
		positions = np.empty((0,2))

		for node in self.nodes:
			positions = np.append(positions, [[node["X"], node["Y"]]], axis=0)

		# Rotation handled before examining the node objects
		if "rotate" in self.globalProperties:
			logger.debug("Rotating Coordinates by {}".format(self.globalProperties["rotate"]))
			self.rotateCoordinates(positions, self.globalProperties["rotate"])


		for i, node in enumerate(self.nodes):
			"""
				addNode default Properties
				(node_name, label=None, shape="rectangle", font_family="Dialog",
                 underlined_text="false", font_style="plain", font_size="12",
                 shape_fill="#FF0000", transparent="false", edge_color="#000000",
                 edge_type="line", edge_width="1.0", height=False, width=False, x=False,
                 y=False, node_type="ShapeNode", UML=False):
			"""

			# Handling Shapes of nodes
			if node["shape"] in ["rectangle", "ellipse", "diamond"] :
				positions[i][0] = positions[i][0] - node["width"]/2.0
				# yed has axis inverted to TiKZ
				# So the formula for y is bit different than standard rotate for cartessian coordinates
				positions[i][1] = -1 * positions[i][1] - node["height"]/2.0

            # Adding all the nodes obtained from CustomTikzListener
			# into graph object
			self.G.add_node(
				node["nodeID"],
				shape=node["shape"],
				label=node["label"],
				x=str(positions[i][0]* self.scalingFactor),
				y=str(positions[i][1]* self.scalingFactor),
				shape_fill=node["shape_fill"],
				edge_color=node["edge_color"],
				height=str(node["height"]* self.scalingFactor ),
				width=str(node["width"]* self.scalingFactor),
				font_size=self.default_fontSize,
				edge_width=node["edge_width"],
				transparent=node["transparent"]
			)

        # Adding all the edge data obtained from CustomTikzListener
		# into graph object
		for edge in self.edges:
			"""
			node1, node2,
			label="", arrowhead="convex", arrowfoot="none",
            color="#000000", line_type="line", width="1.0"
			"""
			self.G.add_edge(
				edge["x"],
				edge["y"],
				arrowhead=edge["arrowhead"],
				arrowfoot=edge["arrowfoot"],
				color=edge["color"],
				width=edge["width"],
				label=edge["label"],
				line_type=edge["line_type"]
			)

		dom = xml.dom.minidom.parseString(self.G.get_graph())
		return dom.toprettyxml()



import io
from os import path
from setuptools import setup, find_packages

this_directory = path.abspath(path.dirname(__file__))
with io.open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    README = f.read()

# This call to setup() does all the work
setup(
    name="tikz2graphml",
    version="1.1.1",
    description="Convert Latex Tikz code into yed graphml",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/ysahil97/tikz-to-yed-graphml",
    author="Harsh Agarwal",
    author_email="harshaga97@gmail.com",
    # license="Apache License 2.0",
    # classifiers=[
    #     "License :: OSI Approved :: MIT License",
    #     "Programming Language :: Python :: 3",
    #     "Programming Language :: Python :: 3.7",
    # ],
    packages=find_packages(exclude=["tests", "*.tests", "*.tests.*", "tests.*"]),
    include_package_data=True,
    install_requires=[
        "colour==0.1.5",
        "numpy==1.16.2",
        "networkx==2.3",
        "matplotlib==3.0.3",
        "pylatexenc==1.4",
        "antlr4-python3-runtime==4.7.2"
    ],
    entry_points={
        "console_scripts": [
            "tikz2graphml=tikz2graphml.__main__:main",
        ]
    }
)



import os
import sys
import logging
from tikz2graphml.parseTikz import ParseTikz
import tkinter
from tkinter import ttk, StringVar, Label, Entry, LEFT, W, Text, END, messagebox
from tkinter.filedialog import askopenfilename, askdirectory

logger = logging.getLogger()

class Gui:
    def __init__(self, window):
        # 'StringVar()' is used to get the instance of input field
        self.inputFileNameVar = StringVar()
        self.outputDirectoryVar = StringVar(value=str(os.path.join(os.getcwd(), 'output')))
        self.scaleVar = StringVar(value='100')
        window.title("Tikz to GraphML convertor")

        startRow = 0

        Label(window ,text = "TikZ to GraphML",font=("Helvetica", 30), width=100).grid(row = startRow, column = 0, sticky=W, ipadx=25, ipady=2,columnspan=10)
        startRow+=1

        # Define Input File Field which is obtained by choosing a file from browsing the file explorer
        Label(window ,text = "Input File Name : ", justify=LEFT).grid(row = startRow,column = 0, sticky=W, ipadx=10, ipady=2)
        ttk.Button(window, text = "Choose File", command = lambda: self.set_input_file_name()).grid(row = startRow,column=3, ipadx=10, ipady=2, padx=10)
        ttk.Entry(window, textvariable = self.inputFileNameVar, width = 70).grid( row = startRow, column = 1, ipadx=1, ipady=2)

        # Define Output Directory to save all the files to that folder
        # (Default) saves it in same directory as input file
        startRow+=1
        Label(window ,text = "Output File Directory : ", justify=LEFT).grid(row = startRow,column = 0, sticky=W, ipadx=10, ipady=2)
        ttk.Button(window, text = "Choose File", command = lambda: self.set_output_directory()).grid(row = startRow,column=3, ipadx=10, ipady=2, padx=10)
        ttk.Entry(window, textvariable = self.outputDirectoryVar, width = 70).grid( row = startRow, column = 1, ipadx=1, ipady=2)

        # Add a scale field to convert resulting graphml files into resonable graphs
        startRow+=1
        Label(window ,text = "Scale : ", justify=LEFT).grid(row = startRow,column = 0, sticky=W, ipadx=10, ipady=2)
        ttk.Entry(window, textvariable = self.scaleVar, width = 70).grid(row = startRow, column = 1, ipadx=1, ipady=2)

        # Convert Tikz files to GraphML files by a button!
        startRow+=1
        ttk.Button(window, text = "Convert", command = lambda: self.convertFile()).grid(row = 20,column = 3, ipadx=1, ipady=1)

    # Set the user inputs on inputFilename and outputDirectoryVar
    def set_input_file_name(self):
        self.inputFileNameVar.set(askopenfilename())
        path, filename = os.path.split(self.inputFileNameVar.get())

    # Set output directory path
    def set_output_directory(self):
        self.outputDirectoryVar.set(askdirectory())

    def get_input_file_name(self):
        return self.inputFileNameVar.get()

    def get_output_directory(self):
        return self.outputDirectoryVar.get()

    def getScale(self):
        return self.scaleVar.get()

    # Main conversion module which tries to convert tikz graphs in graphml
    # format which is stored in specified output directory
    def convertFile(self):
        logger.debug("Converting file")
        logger.debug("InputFile : {}".format(self.get_input_file_name()))
        logger.debug("Output Directory : {}".format(self.get_output_directory()))
        logger.debug("Scale : {}".format(self.getScale()))

        if(not os.path.isfile(self.get_input_file_name())):
            messagebox.showerror("tikz to graphml", "Entered File is Incorrect")
            return

        if not os.path.exists(self.get_output_directory()):
            os.makedirs(self.get_output_directory())

        if(not os.path.isdir(self.get_output_directory())):
            messagebox.showerror("tikz to graphml", "Entered Directory is Incorrect")
            return

        prefix = self.get_input_file_name()
        prefix.replace("\\","/")    # To support os.path.basename in windows systems
        prefix = os.path.basename(os.path.splitext(prefix)[0])

        ParseTikz().run(float(self.getScale()),
            self.get_input_file_name(),
            prefix,
            self.get_output_directory())

        messagebox.showinfo("tikz to graphml", "Conversion Successfull\nFiles stored in {}".format(self.get_output_directory()))



import copy
import math
import logging
import tikz2graphml.parsingUtils as parsingUtils
from tikz2graphml.grammar.TikzListener import TikzListener
from tikz2graphml.grammar.TikzParser import TikzParser
from tikz2graphml.generateGraphml import Graph

logger = logging.getLogger()

# Custom Tikz Listener to augment Tikz Parsing with appropriate custiom actions
class CustomTikzListener(TikzListener):
    def __init__(self, inputFileName: str, outputFileName: str, scalingFactor: float):
        logger.debug("Parsing Tikz Code")
        self.globalProperties = {}
        self.currentNode = {}
        self.currentEdgeList = []
        self.currentEdgeProperty = {}
        self.latestCoordinateX = 0
        self.latestCoordinateY = 0
        self.lastSeenRadius = 1     # Default radius of \draw circle
        self.shapeNodesCoordinates = []
        self.G = Graph(scalingFactor)
        self.inputFileName = inputFileName
        self.outputFileName = outputFileName
        self.nodeIds = {}
        self.numNodeIds = {}

    def exitBegin(self, ctx: TikzParser.BeginContext):
        """
        The exit of begin rule signals the end of graph creation
        Hence, creating the GraphML file here and saving it
        """
        try:
            logger.debug("GlobalProperties : {}".format(self.globalProperties))
            self.G.globalProperties = copy.copy(self.globalProperties)
            graphml = self.G.get_graph().encode("utf-8")
            with open(self.outputFileName, 'wb') as outFile:
                outFile.write(graphml)
            logger.info("Converted Tikz Graph to GraphML")
            logger.info("GraphML File Location: {}\n\n".format(self.outputFileName))
        except Warning as e:
            logger.warn("Error in converting {} - {}".format(self.inputFileName, e))

    def exitGlobalProperties(self, ctx: TikzParser.GlobalPropertiesContext):
        # globalProperties: EVERY (VARIABLE|EXPRESSION) '/.' 'style' '=' '{' properties '}'
        if len(ctx.getTokens(TikzParser.EVERY)) == 1:
            entityForEveryProperty = ctx.getChild(1).getText()
            assert entityForEveryProperty == "node" or \
                entityForEveryProperty == "edge" or \
                entityForEveryProperty == "draw", "Error in parsing {}. Only support \"node,edge,draw\" as ENTITY in \"every <ENTITY>/.style{{}}\"".format(ctx.getText())
            properties = parsingUtils.handleProperties(ctx.getTypedRuleContext(TikzParser.PropertiesContext, 0))
            self.globalProperties.update({entityForEveryProperty: properties})

        # globalProperties: properties
        elif len(ctx.getTypedRuleContexts(TikzParser.PropertiesContext)) == 1:
            properties = parsingUtils.handleProperties(ctx.getTypedRuleContext(TikzParser.PropertiesContext, 0))
            self.globalProperties.update(properties)

    def enterNode(self, ctx: TikzParser.NodeContext):
        """
        Flushing out all node related data before analyzing it
        """
        self.currentNode = {}           # Emptying values before handling a new node
        logger.debug("Parsing Node {}".format(ctx.getText()))
        for k, v in self.globalProperties.items():
            if k == "node":
                self.currentNode.update(v)

    def exitNode(self, ctx: TikzParser.NodeContext):
        """
        Exit of Node rule signals no more node properties
        Hence, storing all those properties in the node dict
        """
        self.currentNode["X"] = self.latestCoordinateX
        self.currentNode["Y"] = self.latestCoordinateY
        if "edge_color" not in self.currentNode:
            self.currentNode["edge_color"] = "black"
        
        # Only send those attributes which are supported
        parsingUtils.filterOutNotSupportedNodeTags(self.currentNode)
        logger.debug("Got Node: {} in input {}".format(self.currentNode, ctx.getText()))
        self.G.addNode(**self.currentNode)

    def exitNodeId(self, ctx: TikzParser.NodeIdContext):
        # nodeID: OPEN_PARANTHESES (VARIABLE|EXPRESSION)? CLOSE_PARANTHESES
        if ctx.getChildCount() == 3:
            """
            Repeated nodeID checking done here
            """
            nodeId = ctx.getChild(1).getText()
            if nodeId in self.nodeIds:
                logger.debug("Got repeated nodeId : {}".format(ctx.getText()))
                self.numNodeIds[nodeId] += 1
                newNodeId = nodeId + "_" + str(self.numNodeIds[nodeId])
                self.currentNode["nodeID"] = newNodeId
                self.nodeIds[nodeId] = newNodeId
            else:
                logger.debug("Got new nodeId : {}".format(ctx.getText()))
                self.nodeIds[nodeId] = nodeId
                self.numNodeIds[nodeId] = 0
                self.currentNode["nodeID"] = nodeId
        else:
            logger.debug("No NodeId present : {}".format(ctx.getText()))
            self.currentNode["nodeID"] = None

    def exitCartesianCoordinates(self, ctx: TikzParser.CartesianCoordinatesContext):
        logger.debug("Cartesian Coordinates {}".format(ctx.getText()))
        self.latestCoordinateX = parsingUtils.handleNumbers(ctx.getChild(1).getText())
        self.latestCoordinateY = parsingUtils.handleNumbers(ctx.getChild(3).getText())
        self.shapeNodesCoordinates.append((self.latestCoordinateX, self.latestCoordinateY))

    def exitPolarCoordinates(self, ctx: TikzParser.PolarCoordinatesContext):
        logger.debug("Polar Coordinates {}".format(ctx.getText()))
        angle = parsingUtils.handleNumbers(ctx.getChild(1).getText())
        r = parsingUtils.handleNumbers(ctx.getChild(3).getText())
        cosA = round(math.cos(math.radians(angle)), 10)
        sinA = round(math.sin(math.radians(angle)), 10)
        self.latestCoordinateX = r * cosA
        self.latestCoordinateY = r * sinA
        logger.debug("Polar Coordinates translated {}".format(self.latestCoordinateX, self.latestCoordinateY))
        self.shapeNodesCoordinates.append((self.latestCoordinateX, self.latestCoordinateY))

    def exitLabel(self, ctx: TikzParser.LabelContext):
        if ctx.LABEL_VARIABLE() is not None:
            label_value = ctx.LABEL_VARIABLE().getText()
            # Stripping off $ at both the ends of the obtained label string in
            # order to get real data contained by the label
            self.currentNode["label"] = parsingUtils.parseLabelValue(label_value)

    def exitRadius(self, ctx: TikzParser.RadiusContext):
        logger.debug("Radius {}".format(ctx.getText()))
        self.lastSeenRadius = ctx.getChild(1).getText()
        self.lastSeenRadius = float(parsingUtils.handleNumbers(self.lastSeenRadius))

    def exitEdgeNode(self, ctx: TikzParser.EdgeNodeContext):
        # edgeNode: OPEN_PARANTHESES (VARIABLE|EXPRESSION) CLOSE_PARANTHESES
        if ctx.getChildCount() == 3:
            if ctx.getChild(1).getText() not in self.nodeIds:
                raise Exception("Node Id in draw operation does not exist : {}".format(ctx.getText()))

            logger.debug("EdgeNode, Original Nodeid: current id -- {}:{}".format(ctx.getChild(1).getText(), self.nodeIds[ctx.getChild(1).getText()]))
            self.currentEdgeList.append(self.nodeIds[ctx.getChild(1).getText()])
        else:
            # edgeNode: coordinates
            coord_x = self.latestCoordinateX
            coord_y = self.latestCoordinateY

            edgeNode = {}
            for k, v in self.globalProperties.items():
                if k == "node":
                    edgeNode.update(v)

            edgeNode["X"] = coord_x
            edgeNode["Y"] = coord_y
            edgeNode["transparent"] = "true"
            # Only send those attributes which are supported
            parsingUtils.filterOutNotSupportedNodeTags(edgeNode)

            newNodeId = self.G.addNode(**edgeNode)
            self.currentEdgeList.append(newNodeId)

    def exitEdgeProperties(self, ctx: TikzParser.EdgePropertiesContext):
        # Inclusion of current edge property into the set of all
        # edge properties for that edge
        if len(ctx.getTypedRuleContexts(TikzParser.PropertiesContext)) == 1:
            edgeProperties = parsingUtils.handleProperties(ctx.getTypedRuleContext(TikzParser.PropertiesContext, 0))
            self.currentEdgeProperty.update(edgeProperties)     # Merging the properties together

    def enterDraw(self, ctx: TikzParser.DrawContext):
        # Emptying data structures before handling a new node/edge
        self.currentEdgeList = []
        self.currentEdgeProperty = {}
        self.currentNode = {}
        self.shapeNodesCoordinates = []
        self.lastSeenRadius = 1     # Default radius of \draw circle

        for k, v in self.globalProperties.items():
            if k == "edge" or k == "draw":
                self.currentEdgeProperty.update(v)

        for k, v in self.globalProperties.items():
            if k == "node" or k == "draw":
                self.currentNode.update(v)

    def exitDraw(self, ctx: TikzParser.DrawContext):
        # \draw shape commands
        if len(ctx.getTypedRuleContexts(TikzParser.CoordinatesContext)) >= 1:

            # \draw[] () node {}
            if len(ctx.getTypedRuleContexts(TikzParser.NodePropertiesContext)) == 1:
                logger.debug("Draw Node Instruction {}".format(ctx.getText()))
                self.currentNode["X"] = self.latestCoordinateX
                self.currentNode["Y"] = self.latestCoordinateY
                logger.debug("NodeProperties : {}".format(self.currentNode))

                # Only send those attributes which are supported
                parsingUtils.filterOutNotSupportedNodeTags(self.currentNode)
                self.G.addNode(**self.currentNode)
            else:
                logger.debug("Draw Shape Instruction {}".format(ctx.getText()))
                node_shape = ctx.getChild(3).getText()
                # handle logic for rectangle drawing
                X = 0
                Y = 0
                height = 0
                width = 0

                if node_shape == 'rectangle':
                    assert len(self.shapeNodesCoordinates) == 2, "Error in parsing {}. Draw shape command has incorrect number of coordinates (!=2)".format(ctx.getText())
                    for i in self.shapeNodesCoordinates:
                        X += i[0]
                        Y += i[1]
                    X /= 2.0
                    Y /= 2.0
                    height = float(abs(int(self.shapeNodesCoordinates[1][1]) - float(self.shapeNodesCoordinates[0][1])))
                    height = 1 if height == 0.0 else height

                    width = float(abs(int(self.shapeNodesCoordinates[1][0]) - float(self.shapeNodesCoordinates[0][0])))
                    width = 1 if width == 0.0 else width

                elif node_shape == 'ellipse':
                    assert len(self.shapeNodesCoordinates) == 2, "Error in parsing {}. Draw shape command has incorrect number of coordinates (!=2)".format(ctx.getText())
                    X = self.shapeNodesCoordinates[0][0]
                    Y = self.shapeNodesCoordinates[0][1]
                    width = float(self.shapeNodesCoordinates[1][0]) * 2
                    height = float(self.shapeNodesCoordinates[1][1]) * 2
                    height = 1 if height == 0.0 else height
                    width = 1 if width == 0.0 else width

                elif node_shape == 'circle':
                    assert len(self.shapeNodesCoordinates) == 1, "Error in parsing {}. Draw circle shape command has incorrect number of coordinates(!=1)".format(ctx.getText())
                    height = float(float(self.lastSeenRadius) * 2)
                    width = float(float(self.lastSeenRadius) * 2)
                    X = float(self.shapeNodesCoordinates[0][0])
                    Y = float(self.shapeNodesCoordinates[0][1])
                    

                else:
                    raise Exception("Error in parsing {}. Currently only support rectangle/circle/ellipse in \draw command".format(ctx.getText()))

                self.G.addNode(X=X, Y=Y, height=height, width=width, shape=node_shape, edge_color="black")

        # \draw line commands
        else:
            logger.debug("Draw Edge(line) Instruction {}".format(ctx.getText()))
            sz = len(self.currentEdgeList)
            arrowFoot = False
            arrowHead = False
            color = "black"
            label = ""
            width = "1"
            line_type = "line"
            if "direction" in self.currentEdgeProperty:
                if self.currentEdgeProperty['direction'] == '->':
                    arrowHead = True
                # For left directed edges, adding the edge nodes in reverse
                elif self.currentEdgeProperty['direction'] == '<-':
                    self.currentEdgeList.reverse()
                    arrowHead = True
                elif self.currentEdgeProperty['direction'] == '-!-':
                    arrowHead = False
                # Only first and last edge should be directed
                elif self.currentEdgeProperty['direction'] == '<->':
                    arrowFoot = True
                    arrowHead = True

            if "fill" in self.currentEdgeProperty:
                color = self.currentEdgeProperty["fill"]

            if "width" in self.currentEdgeProperty:
                width = str(float(self.currentEdgeProperty["width"]) * 0.0352778)

            if "label" in self.currentEdgeProperty:
                label = self.currentEdgeProperty["label"]

            if "line_type" in self.currentEdgeProperty:
                line_type = self.currentEdgeProperty["line_type"]

            for i in range(1, sz, 1):
                nodeX, nodeY = self.currentEdgeList[i-1], self.currentEdgeList[i]
                logger.debug("Adding Edge: {nodeX} {nodeY} {arrowHead} {arrowFoot} {color} {width} {label} {line_type}".format(nodeX=nodeX, nodeY=nodeY, arrowHead=arrowHead, arrowFoot=arrowFoot, color=color, width=width, label=label, line_type=line_type))
                self.G.addEdge(nodeX=nodeX, nodeY=nodeY, arrowHead=arrowHead, arrowFoot=arrowFoot, color=color, width=width, label=label, line_type=line_type)


    def exitNodeProperties(self, ctx: TikzParser.NodePropertiesContext):
        # Inclusion of current node property into the set of all
        # node properties for that node
        if len(ctx.getTypedRuleContexts(TikzParser.PropertiesContext)) == 1:
            nodeProperties = parsingUtils.handleProperties(ctx.getTypedRuleContext(TikzParser.PropertiesContext, 0))
            self.currentNode.update(nodeProperties)



import tkinter
import logging
from tikz2graphml.Gui import Gui

logging.basicConfig(format='%(levelname)-1s : [%(filename)s:%(lineno)d] %(message)s')
logger = logging.getLogger()
logger.setLevel(logging.INFO)

def main():
    # Right now, we use GUI(default) to take in Tikz files and output
    # in specified directory
    window = tkinter.Tk()
    window.geometry("1000x500")

    rows = 0
    while rows < 50:
        window.rowconfigure(rows, weight=2)
        window.columnconfigure(rows, weight=2)
        rows += 1

    gui = Gui(window)
    window.mainloop()

if __name__ == '__main__':
    main()
